@@ -15,27 +15,27 @@ use esp_idf_sys::*;
 use crate::private::common::Newtype;
 use crate::private::cstr::*;
 
-impl From<HttpMethod> for Newtype<(esp_http_client_method_t, ())> {
-    fn from(method: HttpMethod) -> Self {
+impl From<Method> for Newtype<(esp_http_client_method_t, ())> {
+    fn from(method: Method) -> Self {
         Self((
             match method {
-                HttpMethod::Get => esp_http_client_method_t_HTTP_METHOD_GET,
-                HttpMethod::Post => esp_http_client_method_t_HTTP_METHOD_POST,
-                HttpMethod::Delete => esp_http_client_method_t_HTTP_METHOD_DELETE,
-                HttpMethod::Head => esp_http_client_method_t_HTTP_METHOD_HEAD,
-                HttpMethod::Put => esp_http_client_method_t_HTTP_METHOD_PUT,
-                HttpMethod::Options => esp_http_client_method_t_HTTP_METHOD_OPTIONS,
-                HttpMethod::Copy => esp_http_client_method_t_HTTP_METHOD_COPY,
-                HttpMethod::Lock => esp_http_client_method_t_HTTP_METHOD_LOCK,
-                HttpMethod::MkCol => esp_http_client_method_t_HTTP_METHOD_MKCOL,
-                HttpMethod::Move => esp_http_client_method_t_HTTP_METHOD_MOVE,
-                HttpMethod::Propfind => esp_http_client_method_t_HTTP_METHOD_PROPFIND,
-                HttpMethod::Proppatch => esp_http_client_method_t_HTTP_METHOD_PROPPATCH,
-                HttpMethod::Unlock => esp_http_client_method_t_HTTP_METHOD_UNLOCK,
-                HttpMethod::Notify => esp_http_client_method_t_HTTP_METHOD_NOTIFY,
-                HttpMethod::Subscribe => esp_http_client_method_t_HTTP_METHOD_SUBSCRIBE,
-                HttpMethod::Unsubscribe => esp_http_client_method_t_HTTP_METHOD_UNSUBSCRIBE,
-                HttpMethod::Patch => esp_http_client_method_t_HTTP_METHOD_PATCH,
+                Method::Get => esp_http_client_method_t_HTTP_METHOD_GET,
+                Method::Post => esp_http_client_method_t_HTTP_METHOD_POST,
+                Method::Delete => esp_http_client_method_t_HTTP_METHOD_DELETE,
+                Method::Head => esp_http_client_method_t_HTTP_METHOD_HEAD,
+                Method::Put => esp_http_client_method_t_HTTP_METHOD_PUT,
+                Method::Options => esp_http_client_method_t_HTTP_METHOD_OPTIONS,
+                Method::Copy => esp_http_client_method_t_HTTP_METHOD_COPY,
+                Method::Lock => esp_http_client_method_t_HTTP_METHOD_LOCK,
+                Method::MkCol => esp_http_client_method_t_HTTP_METHOD_MKCOL,
+                Method::Move => esp_http_client_method_t_HTTP_METHOD_MOVE,
+                Method::Propfind => esp_http_client_method_t_HTTP_METHOD_PROPFIND,
+                Method::Proppatch => esp_http_client_method_t_HTTP_METHOD_PROPPATCH,
+                Method::Unlock => esp_http_client_method_t_HTTP_METHOD_UNLOCK,
+                Method::Notify => esp_http_client_method_t_HTTP_METHOD_NOTIFY,
+                Method::Subscribe => esp_http_client_method_t_HTTP_METHOD_SUBSCRIBE,
+                Method::Unsubscribe => esp_http_client_method_t_HTTP_METHOD_UNSUBSCRIBE,
+                Method::Patch => esp_http_client_method_t_HTTP_METHOD_PATCH,
                 method => panic!("Method {:?} is not supported", method),
             },
             (),
@@ -127,14 +127,14 @@ impl Drop for EspHttpClient {
     }
 }
 
-impl HttpClient for EspHttpClient {
+impl Client for EspHttpClient {
     type Request<'a> = EspHttpRequest<'a>;
 
     type Error = EspError;
 
     fn request(
         &mut self,
-        method: HttpMethod,
+        method: Method,
         url: impl AsRef<str>,
     ) -> Result<Self::Request<'_>, Self::Error> {
         let c_url = CString::new(url.as_ref()).unwrap();
@@ -149,15 +149,14 @@ impl HttpClient for EspHttpClient {
 
         let follow_redirects = match self.follow_redirects_policy {
             FollowRedirectsPolicy::FollowAll => true,
-            FollowRedirectsPolicy::FollowGetHead => {
-                method == HttpMethod::Get || method == HttpMethod::Head
-            }
+            FollowRedirectsPolicy::FollowGetHead => method == Method::Get || method == Method::Head,
             _ => false,
         };
 
         Ok(EspHttpRequest {
             client: self,
             follow_redirects,
+            size: 0,
         })
     }
 }
@@ -165,40 +164,11 @@ impl HttpClient for EspHttpClient {
 pub struct EspHttpRequest<'a> {
     client: &'a mut EspHttpClient,
     follow_redirects: bool,
+    size: usize,
 }
 
 impl<'a> EspHttpRequest<'a> {
-    fn register_handler(
-        &mut self,
-        handler: impl Fn(&esp_http_client_event_t) -> esp_err_t + 'static,
-    ) {
-        *self.client.event_handler = Some(Box::new(handler));
-    }
-
-    fn deregister_handler(&mut self) {
-        *self.client.event_handler = None;
-    }
-}
-
-impl<'a> HttpRequest<'a> for EspHttpRequest<'a> {
-    type Response<'b> = EspHttpResponse<'b>;
-
-    type Write<'b> = Self;
-
-    type Error = EspError;
-
-    fn send(
-        mut self,
-        size: usize,
-        f: impl FnOnce(&mut Self::Write<'a>) -> Result<(), Self::Error>,
-    ) -> Result<Self::Response<'a>, Self::Error>
-    where
-        Self: Sized,
-    {
-        esp!(unsafe { esp_http_client_open(self.client.raw, size as _) })?;
-
-        f(&mut self)?;
-
+    fn fetch_headers(&mut self) -> Result<BTreeMap<String, String>, EspError> {
         let mut headers = BTreeMap::new();
 
         loop {
@@ -247,7 +217,7 @@ impl<'a> HttpRequest<'a> for EspHttpRequest<'a> {
                         )
                     })?;
                     esp!(unsafe { esp_http_client_set_redirection(self.client.raw) })?;
-                    esp!(unsafe { esp_http_client_open(self.client.raw, size as _) })?;
+                    esp!(unsafe { esp_http_client_open(self.client.raw, self.size as _) })?;
 
                     headers.clear();
 
@@ -258,6 +228,41 @@ impl<'a> HttpRequest<'a> for EspHttpRequest<'a> {
             break;
         }
 
+        Ok(headers)
+    }
+
+    fn register_handler(
+        &mut self,
+        handler: impl Fn(&esp_http_client_event_t) -> esp_err_t + 'static,
+    ) {
+        *self.client.event_handler = Some(Box::new(handler));
+    }
+
+    fn deregister_handler(&mut self) {
+        *self.client.event_handler = None;
+    }
+}
+
+impl<'a> Request<'a> for EspHttpRequest<'a> {
+    type Write<'b> = EspHttpRequest<'b>;
+
+    type Error = EspError;
+
+    fn into_writer(mut self, size: usize) -> Result<Self::Write<'a>, Self::Error> {
+        esp!(unsafe { esp_http_client_open(self.client.raw, size as _) })?;
+
+        self.size = size;
+
+        Ok(self)
+    }
+}
+
+impl<'a> RequestWrite<'a> for EspHttpRequest<'a> {
+    type Response = EspHttpResponse<'a>;
+
+    fn into_response(mut self) -> Result<Self::Response, Self::Error> {
+        let headers = self.fetch_headers()?;
+
         Ok(EspHttpResponse {
             client: self.client,
             headers,
@@ -265,7 +270,7 @@ impl<'a> HttpRequest<'a> for EspHttpRequest<'a> {
     }
 }
 
-impl<'a> HttpSendHeaders<'a> for EspHttpRequest<'a> {
+impl<'a> SendHeaders<'a> for EspHttpRequest<'a> {
     fn set_header<H, V>(&mut self, name: H, value: V) -> &mut Self
     where
         H: Into<Cow<'a, str>>,
@@ -304,24 +309,20 @@ pub struct EspHttpResponse<'a> {
     headers: BTreeMap<String, String>,
 }
 
-impl<'a> HttpResponse<'a> for EspHttpResponse<'a> {
-    type Read<'b> = Self;
+impl<'a> Response for EspHttpResponse<'a> {
+    type Read<'b>
+    where
+        'a: 'b,
+    = &'b EspHttpResponse<'a>;
 
     type Error = EspError;
 
-    fn payload(&mut self) -> &mut Self {
-        self
-    }
-
-    fn into_payload(self) -> Self::Read<'a>
-    where
-        Self: Sized,
-    {
+    fn reader(&self) -> Self::Read<'_> {
         self
     }
 }
 
-impl<'a> HttpHeaders for EspHttpResponse<'a> {
+impl<'a> Headers for EspHttpResponse<'a> {
     fn header(&self, name: impl AsRef<str>) -> Option<Cow<'_, str>> {
         if name.as_ref().eq_ignore_ascii_case("Content-Length") {
             self.content_len().map(|l| Cow::Owned(l.to_string()))
@@ -343,7 +344,7 @@ impl<'a> HttpHeaders for EspHttpResponse<'a> {
     }
 }
 
-impl<'a> HttpStatus for EspHttpResponse<'a> {
+impl<'a> Status for EspHttpResponse<'a> {
     fn status(&self) -> u16 {
         unsafe { esp_http_client_get_status_code(self.client.raw) as _ }
     }
@@ -353,7 +354,7 @@ impl<'a> HttpStatus for EspHttpResponse<'a> {
     }
 }
 
-impl<'a> Read for EspHttpResponse<'a> {
+impl<'a> Read for &EspHttpResponse<'a> {
     type Error = EspError;
 
     fn do_read(&mut self, buf: &mut [u8]) -> Result<usize, Self::Error> {
